/*
 * Bank accounts BST
 */

import tester.*;

/*
 *    +------------------------------+
 * +->| ICompAcct                    |
 * |  +------------------------------+
 * |  | int compareAccts(Acct, Acct) |
 * |  +------------------------------+
 * |
 * |                  +----------------------++
 * |                  |                      | \
 * |                  V                      | |
 * |         +----------------+              | |
 * |         |  ABSTAcct      |              | |
 * |         +----------------+              | |
 * +---------| ICompAcct comp |              | |
 *           +--------+-------+              | |
 *                   / \                     | |
 *                  +---+                    | |
 *                    |                      | |
 *             +-----------------+           | |
 *             |                 |           | |
 *   +----------------+   +----------------+ | |
 *   | LeafAcct       |   | NodeAcct       | | |
 *   +----------------+   +----------------+ | |
 *                      +-| Acct data      | | |
 *                      | | ABSTAcct left  |-+ |
 *                      | | ABSTAcct right |---+
 *                      | +----------------+
 *                      v
 *         +---------------+
 *         | Acct          |
 *         +---------------+
 *         | String owner  |
 *         | String kind   |
 *         | int balance   |
 *         +---------------+
 */

// Represent a Bank Account
class Acct{
    String owner;
    String kind;
    int balance;

    Acct(String owner, String kind, int balance){
        this.owner = owner;
        this.kind = kind;
        this.balance = balance;
    }

    /* Template:
     *   Fields:
     *     ... this.owner ...             -- String
     *     ... this.kind ...              -- String
     *     ... this.balance ...           -- int
     * 
     *   Methods:
     *
     *   Methods for Fields:
     *     ... this.owner.equals(String) ... -- boolean
     *     ... this.kind.equals(String) ...  -- boolean
     */
}

// Represents the comparison of two Accounts
interface ICompAcct{
    // Compare one Account to the other
    public int compareAccts(Acct one, Acct other);
}

// Compare Accounts by their balances
class LowerBalance implements ICompAcct{
    public int compareAccts(Acct one, Acct other){
        return one.balance - other.balance;
    }
}

// Represents a binary search tree of Accounts
abstract class ABSTAcct{
    ICompAcct comp;

    // Compute the total size of this binary search tree
    abstract int size();

    // Add the given Account to this binary search tree
    abstract ABSTAcct add(Acct a);
}

// Represents a leaf in a binary search tree of Accounts
class LeafAcct extends ABSTAcct{
    LeafAcct(ICompAcct comp){
        this.comp = comp;
    }

    // Compute the total size of this binary search tree
    public int size(){
        return 0;
    }

    // Add the given account to this binary search tree
    public ABSTAcct add(Acct b){
        return new NodeAcct(b, this, this);
    }
}

// Represents a node in a binary search tree of Accounts
class NodeAcct extends ABSTAcct{
    Acct data;
    ABSTAcct left;
    ABSTAcct right;

    NodeAcct(Acct data, ABSTAcct left, ABSTAcct right){
        this.data = data;
        this.left = left;
        this.right = right;
        this.comp = this.left.comp;
    }


    /* Template:
     *   Fields:
     *     ... this.comp ...               -- ICompAcct
     *     ... this.data ...               -- Acct
     *     ... this.left ...               -- ABSTAcct
     *     ... this.right ...              -- ABSTAcct
     * 
     *   Methods for Fields:
     *     ... this.comp.compareAccts(Acct, Acct) ... -- int
     * 
     *     ... this.left.size() ...          -- int
     *     ... this.left.add(Acct) ...       -- ABSTAcct
     * 
     *     ... this.right.size() ...         -- int
     *     ... this.right.add(Acct) ...      -- ABSTAcct
     */

    // Compute the total size of this binary search tree
    public int size(){
        return 1 + this.left.size() + this.right.size();
    }

    // Add the given account to this binary search tree
    public ABSTAcct add(Acct b){
        if(this.comp.compareAccts(this.data,b) < 0){
            return new NodeAcct(this.data, this.left, this.right.add(b));
        }else{
            return new NodeAcct(this.data, this.left.add(b), this.right);
        }
    }
}

// Examples and tests for accounts and BSTs of accounts
class AcctBSTExamples{
    AcctBSTExamples(){}

    Acct chk1 = new Acct("JFK", "checking", 4000);
    Acct chk2 = new Acct("LBJ", "checking", 1000);
    Acct sav = new Acct("DDE", "savings", 2000);
    Acct credit = new Acct("RMN", "credit card", 3000);

    ICompAcct compBal = new LowerBalance();

    ABSTAcct leaf = new LeafAcct(this.compBal);

    ABSTAcct node = new NodeAcct(this.sav, this.leaf, this.leaf);

    ABSTAcct nodes2 = new NodeAcct(this.sav,
                                   new NodeAcct(this.chk2, this.leaf, this.leaf),
                                   this.leaf);
    ABSTAcct tree = new NodeAcct(this.sav,
                                 new NodeAcct(this.chk2, this.leaf, this.leaf),
                                 new NodeAcct(this.chk1, this.leaf, this.leaf));

    // Test the compare method in the LowerBalance class
    void testICompAcct(Tester t){
        t.checkExpect(this.compBal.compareAccts(this.chk2, this.chk1) < 0, true);
        t.checkExpect(this.compBal.compareAccts(this.chk1, this.chk2) > 0, true);
    }

    // Test the size method in the BST classes
    void testSize(Tester t){
        t.checkExpect(this.leaf.size(),  0);
        t.checkExpect(this.node.size(),  1);
        t.checkExpect(this.tree.size(),  3);
    }

    // Test the add method in the BST classes
    void testAdd(Tester t){
        t.checkExpect(this.leaf.add(this.sav),  this.node);
        t.checkExpect(this.node.add(this.chk2),  this.nodes2);
        t.checkExpect(this.node.add(this.chk1).add(this.chk2),  this.tree);
    }

}