/*
 * Book BST
 */

import tester.*;

/*
 *    +------------------------------+
 * +->| ICompBook                    |
 * |  +------------------------------+
 * |  | int compareBooks(Book, Book) |
 * |  +------------------------------+
 * |
 * |                  +----------------------++
 * |                  |                      | \
 * |                  V                      | |
 * |         +----------------+              | |
 * |         |  ABSTBook      |              | |
 * |         +----------------+              | |
 * +---------| ICompBook comp |              | |
 *           +--------+-------+              | |
 *                   / \                     | |
 *                  +---+                    | |
 *                    |                      | |
 *             +-----------------+           | |
 *             |                 |           | |
 *   +----------------+   +----------------+ | |
 *   |    LeafBook    |   |    NodeBook    | | |
 *   +----------------+   +----------------+ | |
 *                      +-| Book data      | | |
 *                      | | ABSTBook left  |-+ |
 *                      | | ABSTBook right |---+
 *                      | +----------------+
 *                      V
 *         +---------------+
 *         |     Book      |
 *         +---------------+
 *         | String title  |
 *         | String author |
 *         | int    year   |
 *         +---------------+
 */

// Represents a Book
class Book{
    String title;
    String author;
    int year;

    Book(String title, String author, int year){
        this.title = title;
        this.author = author;
        this.year = year;
    }

    /* Template:
     *   Fields:
     *     ... this.title ...             -- String
     *     ... this.author ...            -- String
     *     ... this.year ...              -- int
     *
     *   Methods
     *
     *   Methods for Fields:
     *     ... this.title.equals(String) ...   -- boolean
     *     ... this.author.equals(String) ...  -- boolean
     */
}

// Represents the comparison of two Books
interface ICompBook{
    // Compare one Book to the other
    public int compareBooks(Book one, Book other);
}

// Compare Books by their year of publication
class EarlierYear implements ICompBook{
    public int compareBooks(Book one, Book other){
        return one.year - other.year;
    }
}

// Represents a binary search tree of Books
abstract class ABSTBook{
    ICompBook comp;

    // Compute the total size of this binary search tree
    abstract int size();

    // Add the given Book to this binary search tree
    abstract ABSTBook add(Book a);
}

// Represents a leaf in a binary search tree of Books
class LeafBook extends ABSTBook{
    LeafBook(ICompBook comp){
        this.comp = comp;
    }

    // compute the total size of this binary search tree
    public int size(){
        return 0;
    }

    // add the given book to this binary search tree
    public ABSTBook add(Book b){
        return new NodeBook(b, this, this);
    }
}

// Represents a node in a binary search tree of Books
class NodeBook extends ABSTBook{
    Book data;
    ABSTBook left;
    ABSTBook right;

    NodeBook(Book data, ABSTBook left, ABSTBook right){
        this.data = data;
        this.left = left;
        this.right = right;
        this.comp = this.left.comp;
    }


    /* Template:
     *   Fields:
     *     ... this.comp ...               -- ICompBook
     *     ... this.data ...               -- Book
     *     ... this.left ...               -- ABSTBook
     *     ... this.right ...              -- ABSTBook
     * 
     *   Methods for Fields:
     *     ... this.comp.compareBooks(Book, Book) ... -- int
     * 
     *     ... this.left.size() ...          -- int
     *     ... this.left.add(Book) ...       -- ABSTBook
     * 
     *     ... this.right.size() ...         -- int
     *     ... this.right.add(Book) ...      -- ABSTBook
     */

    // Compute the total size of this binary search tree
    public int size(){
        return 1 + this.left.size() + this.right.size();
    }

    // Add the given book to this binary search tree
    public ABSTBook add(Book b){
        if(this.comp.compareBooks(this.data,b) < 0){
            return new NodeBook(this.data, this.left, this.right.add(b));
        }else{
            return new NodeBook(this.data, this.left.add(b), this.right);
        }
    }
}

// Examples and tests for books and BSTs of books
class BookBSTExamples{
    BookBSTExamples(){}

    Book cha = new Book("Change in Altitude", "AD", 2009);
    Book oms = new Book("Old Man and the Sea", "EH", 1950);
    Book sar = new Book("Sun Also Rises", "EH", 1954);
    Book htdp = new Book("HtDP", "FFFK", 2000);

    ICompBook compYear = new EarlierYear();

    ABSTBook leaf = new LeafBook(this.compYear);

    ABSTBook node = new NodeBook(this.sar, this.leaf, this.leaf);

    ABSTBook nodes2 = new NodeBook(this.sar,
                                   new NodeBook(this.oms, this.leaf, this.leaf),
                                   this.leaf);
    ABSTBook tree = new NodeBook(this.sar,
                                 new NodeBook(this.oms, this.leaf, this.leaf),
                                 new NodeBook(this.cha, this.leaf, this.leaf));

    // Test the compare method in the EarlierYear class
    void testICompBook(Tester t){
        t.checkExpect(this.compYear.compareBooks(this.oms, this.cha) < 0, true);
        t.checkExpect(this.compYear.compareBooks(this.cha, this.oms) > 0, true);
    }

    // Test the size method in the BST classes
    void testSize(Tester t){
        t.checkExpect(this.leaf.size(),  0);
        t.checkExpect(this.node.size(),  1);
        t.checkExpect(this.tree.size(),  3);
    }

    // Test the add method in the BST classes
    void testAdd(Tester t){
        t.checkExpect(this.leaf.add(this.sar),  this.node);
        t.checkExpect(this.node.add(this.oms),  this.nodes2);
        t.checkExpect(this.node.add(this.cha).add(this.oms),  this.tree);
    }

}